import string
import gpio

# ------------------------------------------------------- #
#                         Requests                        #
# ------------------------------------------------------- #
class NBIoTRequestType
  static var DEBUG = 0
  static var MQTT_CONN = 1
  static var MQTT_PUB = 2
  static var MQTT_DISC = 3
  static var COAP = 4
  static var HTTP = 5
  static var NTP = 6
  static var PSM_ENABLE = 7
end

class NBIoTRequest
  var request_type
  var callback

  def init(request_type, callback)
    self.request_type = request_type
    self.callback = callback
  end
end

class NBIoTDebugRequest : NBIoTRequest
  var cmd
  var rsp
  var retries

  def init(cmd, rsp, retries)
    super(self).init(NBIoTRequestType.DEBUG)
    self.cmd = cmd
    self.rsp = rsp
    self.retries = retries
  end
end

class NBIoTMQTTConnectRequest : NBIoTRequest
  var host
  var port
  var client
  var username
  var password

  def init(host, port, client, username, password, callback)
    super(self).init(NBIoTRequestType.MQTT_CONN, callback)
    self.host = host
    self.port = port
    self.client = client
    self.username = username
    self.password = password
  end
end

class NBIoTMQTTPublishRequest : NBIoTRequest
  var topic
  var payload

  def init(topic, payload, callback)
    super(self).init(NBIoTRequestType.MQTT_PUB, callback)
    self.topic = topic
    self.payload = payload
  end
end

class NBIoTCOAPRequest : NBIoTRequest
  var host
  var port
  var paths
  var method
  var queries
  var topic
  var payload

  def init(host, port, paths, method, queries, payload, callback)
    super(self).init(NBIoTRequestType.COAP, callback)
    self.host = host
    self.port = port
    self.paths = paths
    self.method = method
    self.queries = queries
    self.payload = payload
  end
end

class NBIoTHTTPRequest : NBIoTRequest
  var url
  var method
  var payload

  def init(url, method, payload, callback)
    super(self).init(NBIoTRequestType.HTTP, callback)
    self.url = url
    self.method = method
    self.payload = payload
  end
end

# ------------------------------------------------------- #
#                        Procedures                       #
# ------------------------------------------------------- #

class NBIoTCommand
  var cmd
  var rsp
  var retries

  def init(cmd, rsp, retries)
    self.cmd = cmd
    self.rsp = rsp
    self.retries = retries

    if self.retries == nil
      self.retries = 0
    end
  end
end

class NBIoTReadDriver
  var ser
  var buf
  var flush
  var printed

  def fetch_and_flush()
    self.flush = true
    self.printed = false
    return self.buf
  end

  def fast_loop()
    if self.flush
      self.buf = ''
      self.flush = false
    end

    var bytes_available = self.ser.available()

    if (size(self.buf) + bytes_available) >= 2048
      self.buf = self.buf[(size(self.buf) + bytes_available) - 2048 ..]
    end

    if bytes_available != 0
      self.buf += self.ser.read().asstring()
    end
    
    # todo remove
    if string.find(self.buf, 'xcx') >= 0 && !self.printed
      tasmota.log('------time stop------', 2) # todo remove
      # tasmota.log('buf size: ' + str(size(self.buf)), 2)
      # tasmota.log(self.buf, 2)
      self.printed = true
    end
  end
  
  def init(ser)
    self.ser = ser
    self.buf = ''
    self.flush = false
    self.printed = false

    tasmota.add_fast_loop(/-> self.fast_loop()) # todo deactivate when psm enabled
  end
end

class NBIoTProcedure
  var ser
  var read_drv
  var request
  var response
  var debug
  var done
  var aborted
  var rsp_awaiting
  var cmd_in_process

  def init(ser, read_drv, request)
    self.ser = ser
    self.read_drv = read_drv
    self.request = request
    self.response = nil
    self.debug = false
    self.done = false
    self.aborted = false
    self.rsp_awaiting = false
    self.cmd_in_process = nil
  end

  def is_done()
    return self.done
  end

  def is_aborted()
    return self.aborted
  end

  def send_cmd()
    if self.cmd_in_process.cmd != nil
      self.ser.write(bytes().fromstring(self.cmd_in_process.cmd))

      tasmota.log(string.format('NBT: Sending command \'%s\'', string.replace(self.cmd_in_process.cmd, '\r\n', '')), 2)
    end

    self.rsp_awaiting = true
  end

  def read_rsp_contains_expected_rsp(return_rsp)
    if self.rsp_awaiting
      var rsp = self.read_drv.fetch_and_flush()

      if self.debug
        print(rsp)
      end

      if string.find(rsp, 'OK') >= 0
        tasmota.log('------time start------', 2) # todo remove
      end

      if string.find(rsp, self.cmd_in_process.rsp) >= 0
        self.rsp_awaiting = false

        tasmota.log(string.format('NBT: Received \'%s\'', self.cmd_in_process.rsp), 2)
        
        if return_rsp
          return rsp
        end

        return true
      else
        if self.cmd_in_process.retries > 0
          self.cmd_in_process.retries -= 1
        end
      end
    end

    if return_rsp
      return nil
    else
      return false
    end
  end

  def fetch_read_rsp_if_contains_expected_rsp_or_send()
    var rsp = self.read_rsp_contains_expected_rsp(true)

    if rsp != nil
      return rsp
    else
      self.send_cmd()

      return nil
    end
  end

  def read_rsp_contains_expected_rsp_or_send()
    if self.read_rsp_contains_expected_rsp(false)
      return true
    else
      self.send_cmd()

      return false
    end
  end

  def retries_exceeded()
    return self.cmd_in_process.retries < 1
  end

  def execute() end
end

class NBIoTDebugProcedure : NBIoTProcedure
  var cmd_debug

  def init(ser, read_drv, request)
    super(self).init(ser, read_drv, request)

    self.debug = true
    self.cmd_debug = NBIoTCommand(request.cmd + '\r\n', request.rsp, request.retries)
  end

  def execute()
    if self.cmd_in_process == nil
      self.cmd_in_process = self.cmd_debug
    end

    self.done = self.read_rsp_contains_expected_rsp_or_send()
    self.aborted = self.retries_exceeded()
  end
end

class NBIoTResetProcedure : NBIoTProcedure
  var cmd_reset
  var cmd_at

  def init(ser, read_drv)
    super(self).init(ser, read_drv)

    self.cmd_reset = NBIoTCommand('AT+QRST=1\r\n', 'RDY', 10)
    self.cmd_at = NBIoTCommand('AT\r\n', 'OK')
  end

  def execute()
    if self.cmd_in_process == nil
      self.cmd_in_process = self.cmd_reset
    end

    if self.cmd_in_process == self.cmd_reset && self.read_rsp_contains_expected_rsp_or_send()
      self.cmd_in_process = self.cmd_at
    elif self.cmd_in_process == self.cmd_at
      self.done = self.read_rsp_contains_expected_rsp_or_send()
    end

    self.aborted = self.retries_exceeded()
  end
end

class NBIoTEnablePSMProcedure : NBIoTProcedure
  var cmd_register
  var cmd_enable_psm
  var cmd_enable_sleep

  def init(ser, read_drv, request)
    super(self).init(ser, read_drv, request)

    self.cmd_register = NBIoTCommand('AT+CEREG=5\r\n', 'OK', 10)
    self.cmd_enable_psm = NBIoTCommand('AT+CPSMS=1,,,\"01011111\",\"00000001\"\r\n', 'OK', 10)
    self.cmd_enable_sleep = NBIoTCommand('AT+QSCLK=1\r\n', 'OK', 10)
  end

  def execute()
    if self.cmd_in_process == nil
      self.cmd_in_process = self.cmd_register
    end

    if self.cmd_in_process == self.cmd_register && self.read_rsp_contains_expected_rsp_or_send()
      self.cmd_in_process = self.cmd_enable_psm
    elif self.cmd_in_process == self.cmd_enable_psm && self.read_rsp_contains_expected_rsp_or_send()
      self.cmd_in_process = self.cmd_enable_sleep
    elif self.cmd_in_process == self.cmd_enable_sleep
      self.done = self.read_rsp_contains_expected_rsp_or_send()
    end

    self.aborted = self.retries_exceeded()
  end
end

class NBIoTDisablePSMProcedure : NBIoTProcedure
  var cmd_register
  var cmd_disable_psm
  var cmd_disable_sleep

  def init(ser, read_drv)
    super(self).init(ser, read_drv)

    self.cmd_register = NBIoTCommand('AT+CEREG=1\r\n', 'OK')
    self.cmd_disable_psm = NBIoTCommand('AT+CPSMS=0\r\n', 'OK')
    self.cmd_disable_sleep = NBIoTCommand('AT+QSCLK=0\r\n', 'OK')
  end

  def execute()
    if self.cmd_in_process == nil
      self.cmd_in_process = self.cmd_register
    end

    if self.cmd_in_process == self.cmd_register && self.read_rsp_contains_expected_rsp_or_send()
      self.cmd_in_process = self.cmd_disable_psm
    elif self.cmd_in_process == self.cmd_disable_psm && self.read_rsp_contains_expected_rsp_or_send()
      self.cmd_in_process = self.cmd_disable_sleep
    elif self.cmd_in_process == self.cmd_disable_sleep
      self.done = self.read_rsp_contains_expected_rsp_or_send()
    end
  end
end

class NBIoTMQTTConnectProcedure : NBIoTProcedure
  var cmd_open
  var cmd_conn

  def init(ser, read_drv, request)
    super(self).init(ser, read_drv, request)

    var cmd = 'AT+QMTOPEN=0,\"%s\",%s\r\n'
    self.cmd_open = NBIoTCommand(string.format(cmd, request.host, str(request.port)), '+QMTOPEN: 0,0', 15)

    cmd = 'AT+QMTCONN=0,\"%s\",\"%s\",\"%s\"\r\n'
    self.cmd_conn = NBIoTCommand(string.format(cmd, request.client, request.username, request.password), '+QMTCONN: 0,0,0', 15)
  end

  def execute()
    if self.cmd_in_process == nil
      self.cmd_in_process = self.cmd_open
    end
    
    if self.cmd_in_process == self.cmd_open && self.read_rsp_contains_expected_rsp_or_send()
      self.cmd_in_process = self.cmd_conn
    elif self.cmd_in_process == self.cmd_conn
      self.done = self.read_rsp_contains_expected_rsp_or_send()
    end

    self.aborted = self.retries_exceeded()
  end
end

class NBIoTMQTTPublishProcedure : NBIoTProcedure
  var cmd_conn
  var cmd_pub

  def init(ser, read_drv, request)
    super(self).init(ser, read_drv, request)

    self.cmd_conn = NBIoTCommand('AT+QMTCONN?\r\n', '+QMTCONN: 0,3', 10)

    var cmd = 'AT+QMTPUB=0,0,0,0,\"%s\",\"%s\"\r\n'
    self.cmd_pub = NBIoTCommand(string.format(cmd, request.topic, request.payload), '+QMTPUB: 0,0,0', 10)
  end

  def execute()
    if self.cmd_in_process == nil
      self.cmd_in_process = self.cmd_conn
    end

    if self.cmd_in_process == self.cmd_conn && self.read_rsp_contains_expected_rsp_or_send()
      self.cmd_in_process = self.cmd_pub
    elif self.cmd_in_process == self.cmd_pub
      self.done = self.read_rsp_contains_expected_rsp_or_send()
    end

    self.aborted = self.retries_exceeded()
  end
end

class NBIoTMQTTDisconnectProcedure : NBIoTProcedure
  var cmd_close

  def init(ser, read_drv, request)
    super(self).init(ser, read_drv, request)

    self.cmd_close = NBIoTCommand('AT+QMTCLOSE=0\r\n', '+QMTCLOSE: 0,0', 10)
  end

  def execute()
    if self.cmd_in_process == nil
      self.cmd_in_process = self.cmd_close
    end

    self.done = self.read_rsp_contains_expected_rsp_or_send()
    self.aborted = self.retries_exceeded()
  end
end

class NBIoTNTPProcedure : NBIoTProcedure
  var cmd_ntp

  def init(ser, read_drv, request)
    super(self).init(ser, read_drv, request)

    self.cmd_ntp = NBIoTCommand('AT+QNTP=1,\"0.at.pool.ntp.org\"\r\n', '+QNTP: 0', 10)
  end

  def extract_timestamp(rsp)
    var rsp_args = string.split(rsp, '\"')
    var timestamp = nil

    for rsp_arg : rsp_args
      timestamp = tasmota.strptime(rsp_arg, "%y/%m/%d,%H:%M:%S")

      if timestamp != nil
        break
      end
    end

    return timestamp['epoch'] + 3600 # UTC + 1
  end

  def execute()
    if self.cmd_in_process == nil
      self.cmd_in_process = self.cmd_ntp
    end

    var rsp = self.fetch_read_rsp_if_contains_expected_rsp_or_send()

    if rsp != nil
      self.response = self.extract_timestamp(rsp)
      self.done = true
    end

    self.aborted = self.retries_exceeded()
  end
end

class NBIoTCOAPGetProcedure : NBIoTProcedure
  var cmd_create
  var cmd_url
  var cmd_send
  var cmd_del

  def init(ser, read_drv, request)
    super(self).init(ser, read_drv, request)

    self.cmd_create = NBIoTCommand('AT+QCOAPCREATE=1,0\r\n', 'OK', 10)
    
    var cmd = string.format('AT+QCOAPOPTION=%i', request.paths.size() + request.queries.size())
  
    for path : request.paths
      cmd += string.format(',11,\"%s\"', path)
    end

    for query : request.queries
      cmd += string.format(',11,\"%s\"', query)
    end
 
    cmd += '\r\n'

    self.cmd_url = NBIoTCommand(cmd, 'OK', 10)
    
    cmd = 'AT+QCOAPSEND=0,1,\"%s\",%s,0\r\n'
    self.cmd_send = NBIoTCommand(string.format(cmd, request.host, str(request.port)), '+QCOAPSEND: 0', 1)
    self.cmd_del = NBIoTCommand('AT+QCOAPDEL\r\n', 'OK', 10)
  end

  def execute()
    if self.cmd_in_process == nil
      self.cmd_in_process = self.cmd_create
    end
    
    if self.cmd_in_process == self.cmd_create && self.read_rsp_contains_expected_rsp_or_send()
      self.cmd_in_process = self.cmd_url
    elif self.cmd_in_process == self.cmd_url && self.read_rsp_contains_expected_rsp_or_send()
      self.cmd_in_process = self.cmd_send

      if !self.done
        self.read_rsp_contains_expected_rsp_or_send()
        self.ser.flush()
      end

      self.done = true
    end
    
    self.aborted = self.retries_exceeded()
  end
end

class NBIoTHTTPGetProcedure : NBIoTProcedure
  var cmd_url
  var cmd_input
  var cmd_get
  var cmd_wait
  var cmd_read
  var payload_size

  def init(ser, read_drv, request)
    super(self).init(ser, read_drv, request)

    var url = request.url

    if size(url) > 0 && url[size(url)-1] != '/'
      url += '/'
    end

    self.cmd_url = NBIoTCommand(string.format('AT+QHTTPURL=%i,15\r\n', size(url)), '>', 30)
    self.cmd_input = NBIoTCommand(string.format('%s\r\n', url), 'OK', 30)
    self.cmd_get = NBIoTCommand('AT+QHTTPGET=15\r\n', 'OK', 50)
    self.cmd_wait = NBIoTCommand(nil, '+QHTTPGET: 0,', 100)
    self.cmd_read = NBIoTCommand('AT+QHTTPREAD=256\r\n', '+QHTTPREAD:', 100)

    self.payload_size = 0
    self.response = ''
  end
  
  def execute()
    if self.cmd_in_process == nil
      self.cmd_in_process = self.cmd_url
    end
    
    if self.cmd_in_process == self.cmd_url && self.read_rsp_contains_expected_rsp_or_send()
      self.cmd_in_process = self.cmd_input
    elif self.cmd_in_process == self.cmd_input && self.read_rsp_contains_expected_rsp_or_send()
      self.cmd_in_process = self.cmd_get
    elif self.cmd_in_process == self.cmd_get && self.read_rsp_contains_expected_rsp_or_send()
      self.cmd_in_process = self.cmd_wait
    elif self.cmd_in_process == self.cmd_wait
      var rsp = self.fetch_read_rsp_if_contains_expected_rsp_or_send()

      if rsp != nil
        var params = string.split(rsp, ',')
        var payload_size = params[params.size()-1]

        self.payload_size = int(string.replace(payload_size, '\r\n', '')) + 256
        self.cmd_in_process = self.cmd_read
      end
    elif self.cmd_in_process == self.cmd_read
      var rsp = self.fetch_read_rsp_if_contains_expected_rsp_or_send()

      if rsp != nil
        self.response += rsp # todo filter
        # print('rsp ' + rsp)
        self.done = size(self.response) > self.payload_size

        if self.done
          # print(self.response)
        end
      end
    end
    
    self.aborted = self.retries_exceeded()
  end
end

# ------------------------------------------------------- #
#                       Driver State                      #
# ------------------------------------------------------- #

class NBIoTDriverState
  static var INIT = 0
  static var PSM = 1
  static var RESET = 2
  static var DISABLE_PSM = 3
  static var READY = 4
  static var BUSY = 5

  static def tostring(state)
    if state == NBIoTDriverState.INIT
      return 'INIT'
    elif state == NBIoTDriverState.PSM
      return 'PSM'
    elif state == NBIoTDriverState.RESET
      return 'RESET'
    elif state == NBIoTDriverState.DISABLE_PSM
      return 'DISABLE_PSM'
    elif state == NBIoTDriverState.READY
      return 'READY'
    elif state == NBIoTDriverState.BUSY
      return 'BUSY'
    end
  end
end

# ------------------------------------------------------- #
#                          Driver                         #
# ------------------------------------------------------- #

class NBIoTDriver
  var ser
  var psm_eint
  var read_drv
  var state
  var request_queue
  var procedure
  var psm_disabled

  def init(rx, tx, psm_eint)
    self.ser = serial(rx, tx, 115200, serial.SERIAL_8N1)
    self.psm_eint = psm_eint

    self.read_drv = NBIoTReadDriver(self.ser)

    self.state = NBIoTDriverState.INIT
    self.request_queue = []
    self.procedure = nil
    self.psm_disabled = false
  end

  def queue_request(request)
    if self.state != NBIoTDriverState.PSM
      self.request_queue.push(request)
          
      return true
    end 

    return false
  end

  def disable_psm()
    if self.state == NBIoTDriverState.PSM && !self.psm_disabled
      self.psm_disabled = true
      
      return true
    end 

    return false
  end

  def start_procedure(procedure)
    self.procedure = procedure
  end

  def finish_procedure(aborted)
    if self.procedure.request != nil && self.procedure.request.callback != nil
      var error_code = 0

      if aborted
        error_code = -1
      end

      self.procedure.request.callback(error_code, self.procedure.response)
    end

    self.procedure = nil
  end

  def next_state(state)
    var current_state = NBIoTDriverState.tostring(self.state)
    var next_state = NBIoTDriverState.tostring(state)

    tasmota.log(string.format('NBT: Transitioning from state %s to %s', current_state, next_state), 2)

    self.state = state
  end

  def set_psm_pin_high()
    tasmota.log(string.format('NBT: Setting PSM_EINT pin to %i', 1), 2)

    gpio.digital_write(self.psm_eint, 1)
  end

  def set_psm_pin_low()
    tasmota.log(string.format('NBT: Setting PSM_EINT pin to %i', 0), 2)

    gpio.digital_write(self.psm_eint, 0)
  end

  def wake_up()
    self.set_psm_pin_low()

    tasmota.set_timer(150, /-> self.set_psm_pin_high())
  end

  def every_250ms()
    if self.state == NBIoTDriverState.PSM
      if self.psm_disabled
        self.next_state(NBIoTDriverState.INIT)
      end
    elif self.state == NBIoTDriverState.INIT
      self.set_psm_pin_high()
      self.next_state(NBIoTDriverState.RESET)

      tasmota.set_timer(150, /-> self.wake_up())
    elif self.state == NBIoTDriverState.RESET
      if self.procedure == nil
        self.start_procedure(NBIoTResetProcedure(self.ser, self.read_drv))
      end

      self.procedure.execute()

      if self.procedure.is_done()
        self.finish_procedure(false)

        if self.psm_disabled
          self.next_state(NBIoTDriverState.READY)
        else
          self.next_state(NBIoTDriverState.DISABLE_PSM)
        end
      elif self.procedure.is_aborted()
        self.next_state(NBIoTDriverState.INIT)
      end
    elif self.state == NBIoTDriverState.DISABLE_PSM
      if self.procedure == nil
        self.start_procedure(NBIoTDisablePSMProcedure(self.ser, self.read_drv))
      end

      self.procedure.execute()

      if self.procedure.is_done()
        self.psm_disabled = true
        self.finish_procedure(false)
        self.next_state(NBIoTDriverState.READY)
      end
    elif self.state == NBIoTDriverState.READY
      if self.request_queue.size() > 0
        var request = self.request_queue[0]
        self.request_queue.remove(0)

        var procedure = nil

        if request.request_type == NBIoTRequestType.DEBUG
          procedure = NBIoTDebugProcedure(self.ser, self.read_drv, request)
        elif request.request_type == NBIoTRequestType.MQTT_CONN
          procedure = NBIoTMQTTConnectProcedure(self.ser, self.read_drv, request)
        elif request.request_type == NBIoTRequestType.MQTT_PUB
          procedure = NBIoTMQTTPublishProcedure(self.ser, self.read_drv, request)
        elif request.request_type == NBIoTRequestType.MQTT_DISC
          procedure = NBIoTMQTTDisconnectProcedure(self.ser, self.read_drv, request)
        elif request.request_type == NBIoTRequestType.NTP
          procedure = NBIoTNTPProcedure(self.ser, self.read_drv, request)
        elif request.request_type == NBIoTRequestType.COAP && request.method == 'GET'
          procedure = NBIoTCOAPGetProcedure(self.ser, self.read_drv, request)
        elif request.request_type == NBIoTRequestType.HTTP && request.method == 'GET'
          procedure = NBIoTHTTPGetProcedure(self.ser, self.read_drv, request)
        elif request.request_type == NBIoTRequestType.PSM_ENABLE
          procedure = NBIoTEnablePSMProcedure(self.ser, self.read_drv, request)
        else
          tasmota.log(string.format('NBT: Request with type %i not supported, discarding request', request.request_type), 2)
        end

        if procedure != nil
          self.start_procedure(procedure)
          self.next_state(NBIoTDriverState.BUSY)
          self.procedure.execute()
        end
      end
    elif self.state == NBIoTDriverState.BUSY
      if self.procedure.is_aborted() || self.procedure.is_done()
        if self.procedure.is_aborted()
          var cmd = string.replace(self.procedure.cmd_in_process.cmd, '\r\n', '')

          tasmota.log(string.format('NBT: Exceeded retries at command %s, discarding request', cmd), 2)
        end

        if self.procedure.request.request_type == NBIoTRequestType.PSM_ENABLE && self.procedure.is_done()
          self.psm_disabled = false
        end

        self.finish_procedure(self.procedure.is_aborted())

        if self.psm_disabled
          self.next_state(NBIoTDriverState.READY)
        else
          self.set_psm_pin_low()
          self.next_state(NBIoTDriverState.PSM)
        end
      else
        self.procedure.execute()
      end
    else
      tasmota.log('NBT: Invalid driver state, stopping driver', 2)
      tasmota.remove_driver(self)
    end
  end
end

# ------------------------------------------------------- #
#                       nbiot module                      #
# ------------------------------------------------------- #

var nbiot = module('nbiot')

nbiot.init = def (m)
  class nbiot
    var _driver
    var _started

    def init()
      self._driver = NBIoTDriver(18, 23, 16)
      self._started = false
    end

    def start()
      if !self._started
        tasmota.add_driver(self._driver)
      end
    end

    def enable_power_save_mode(callback)
      var request = NBIoTRequest(NBIoTRequestType.PSM_ENABLE, callback)

      return self._driver.queue_request(request)
    end

    def disable_power_save_mode()
      return self._driver.disable_psm()
    end

    def get_timestamp(callback)
      var request = NBIoTRequest(NBIoTRequestType.NTP, callback)

      return self._driver.queue_request(request)
    end

    def mqtt_connect(host, port, client, username, password, callback)
      var request = NBIoTMQTTConnectRequest(host, port, client, username, password, callback)

      return self._driver.queue_request(request)
    end

    def mqtt_disconnect(callback)
      var request = NBIoTRequest(NBIoTRequestType.MQTT_DISC, callback)

      return self._driver.queue_request(request)
    end

    def mqtt_publish(topic, payload, callback)
      var request = NBIoTMQTTPublishRequest(topic, payload, callback)

      return self._driver.queue_request(request)
    end

    def http_get(url, callback)
      var request = NBIoTHTTPRequest(url, 'GET', nil, callback)

      return self._driver.queue_request(request)
    end

    def coap_get(host, port, paths, queries, callback)
      if type(paths) != type([]) || type(queries) != type([])
        return false
      else
        var request = NBIoTCOAPRequest(host, port, paths, 'GET', queries, nil, callback)

        return self._driver.queue_request(request)
      end
    end

    def coap_post(host, port, paths, payload, callback)
      if type(paths) != type([])
        return false
      else
        var request = NBIoTCOAPRequest(host, port, paths, 'POST', [], payload, callback)

        return self._driver.queue_request(request)
      end
    end

    def debug_send(cmd, rsp, retries)
      if retries == nil
        retries = 0
      end
      
      var request = NBIoTDebugRequest(cmd, rsp, retries)

      return self._driver.queue_request(request)
    end
  end

  return nbiot()
end

return nbiot