import mqtt
import json
import otaconf
import persist
import mesh
import path
import string
import re

var otaconf_finished = false

var _otaconf_timer_id = 'otaconf_timer'
var _file_transfer_timer_id = 'file_transfer_timer'
var _wifi_used = tasmota.cmd('Wifi')['Wifi'] == 'ON'
var _mesh_scan_retries = 3
var _configuration_uuid = nil
var _configuration_files = nil
var _configuration_commands = nil
var _own_file = nil
var _current_chunk = nil
var _current_chunk_id = 0
var _last_chunk_id = 0
var _file_uuid_in_process = nil
var _file_in_process = nil
var _devices_to_receive_file = {}
var _file_send_retries = 3
var _own_command = nil

var _CHUNK_HEADER_SIZE = 12 + 3


def _set_ota_conf_interval(cmd, idx, payload)
  if otaconf.set_poll_interval(payload)
    tasmota.resp_cmnd_done()
  else
    tasmota.resp_cmnd_failed()
  end
end

def _set_ota_server_url(cmd, idx, payload)
  otaconf.set_server_url(payload)
  tasmota.resp_cmnd_done()
end

def _otaconf_timeout()
  otaconf_finished = true

  tasmota.log('OTA: configuration timed out', 1)
end

def _otaconf_failed()
  otaconf_finished = true

  tasmota.log('OTA: configuration failed', 1)
end

def _finish_otaconf(early_exit)
  if !early_exit
    persist.otaconf_last_poll = tasmota.rtc()['local']
    persist.save()
  end

  otaconf_finished = true

  if _own_command != nil
    tasmota.cmd(_own_command)
  end

  tasmota.log('OTA: configuration finished', 2)
end

def _config_status_updated()
  tasmota.remove_timer(_otaconf_timer_id)

  _finish_otaconf(false)
end

def _update_config_status()
  if _configuration_uuid != nil
    otaconf.update_config_status(_configuration_uuid, _config_status_updated)
    tasmota.set_timer(3000, _otaconf_timeout, _otaconf_timer_id)
  else
    _otaconf_failed()
  end
end

def _execute_commands()
  while _configuration_commands.size() > 0
    var configuration_command = _configuration_commands[0]
    var device_macs = configuration_command['macs']
    var command = configuration_command['command']

    for device_mac : device_macs      
      if device_mac == otaconf.get_mac()
        _own_command = command
      else
        tasmota.log('OTA: remotely executing command on device: ' + device_mac, 2)
      
        var payload = {'action': 'EXECUTE_COMMAND', 'command': command}
        mesh.send(otaconf.format_mac(device_mac), json.dump(payload))
      end
    end

    _configuration_commands.remove(0)
  end

  _update_config_status()
end

def _send_file_chunk()
  tasmota.remove_timer(_file_transfer_timer_id)

  if _configuration_files.size() > 0
    var file = _configuration_files[0]
    var uuid = file['uuid']

    if !_devices_to_receive_file.contains(uuid) || _devices_to_receive_file[uuid].size() == 0
      _configuration_files.remove(0)

      _send_file_chunk()   
      return
    end

    var mac = _devices_to_receive_file[uuid][0]

    if _file_in_process == nil
      if !path.exists(uuid)
        tasmota.log('OTA: file: ' + uuid + ' not found - skip', 2)

        _configuration_files.remove(0)

        _send_file_chunk()
        return
      end

      tasmota.log('OTA: opening file: ' + uuid, 2)

      _file_in_process = open(uuid)
      _file_send_retries = 3
      _current_chunk_id = 0
      _last_chunk_id = 0
      _current_chunk = _file_in_process.read(mesh.MAX_LEN - _CHUNK_HEADER_SIZE - 1)

      tasmota.log('OTA: sending chunk: ' + str(_current_chunk_id) + ' to: ' + mac, 2)

      var payload = uuid + string.format("%03i", _current_chunk_id) + _current_chunk
      print('send payload: ' + payload)
      print(size(payload))
      print(mesh.MAX_LEN)

      mesh.send(mac, payload)
    else
      if _current_chunk_id == _last_chunk_id
        _file_send_retries -= 1

        if _file_send_retries > 0
          tasmota.log('OTA: retry sending chunk: ' + str(_current_chunk_id) + ' to: ' + mac + ', retries left: ' + str(_file_send_retries), 2)

          var payload = uuid + string.format("%03i", _current_chunk_id) + _current_chunk
          print('send payload: ' + payload)
  
          mesh.send(mac, payload)
        else
          _devices_to_receive_file[uuid].remove(0)

          _file_in_process.close()
          _file_in_process = nil

          tasmota.log('OTA: retries exceeded: ' + str(_current_chunk_id) + ' to: ' + mac, 2)
        end
      else
        _current_chunk = _file_in_process.read(mesh.MAX_LEN - _CHUNK_HEADER_SIZE - 1)

        if size(_current_chunk) == 0
          _file_in_process.close()
          _file_in_process = nil

          _devices_to_receive_file[uuid].remove(0)

          tasmota.log('OTA: finished transfering file: ' + uuid + ' to: ' + mac, 2)
        else
          tasmota.log('OTA: sending chunk: ' + str(_current_chunk_id) + ' to: ' + mac, 2)

          var payload = uuid + string.format("%03i", _current_chunk_id) + _current_chunk
          print('send payload: ' + payload)

          _last_chunk_id = _current_chunk_id
  
          mesh.send(mac, payload)
        end
      end
    end

    tasmota.set_timer(3000, _send_file_chunk, _file_transfer_timer_id)
  else
    _execute_commands()
  end
end

def _initiate_file_transfer()
  if _configuration_files.size() > 0
    for file : _configuration_files
      var device_macs = file['macs']
      var uuid = file['uuid']
      var name = file['name']

      tasmota.log('OTA: processing file: ' + uuid, 2)

      var f = open(uuid)
      var num_chunks = size(f) / (mesh.MAX_LEN - _CHUNK_HEADER_SIZE - 1)

      if num_chunks == 0 && size(f) > 0
        num_chunks = 1
      end

      if size(f) % (mesh.MAX_LEN - _CHUNK_HEADER_SIZE - 1) != 0
        num_chunks + 1
      end

      f.close()
      
      for device_mac : device_macs      
        if device_mac == otaconf.get_mac()
          # todo own file
          _own_file = file
        else
          tasmota.log('OTA: initiating file transfer with device: ' + device_mac, 2)
        
          var payload = {'action': 'TRANSFER_FILE', 'name': name, 'uuid': uuid, 'num_chunks': num_chunks}

          print('send payload: ' + json.dump(payload))

          mesh.send(otaconf.format_mac(device_mac), json.dump(payload))
        end
      end
    end

    tasmota.set_timer(5000, _send_file_chunk, _file_transfer_timer_id)
  else
    _execute_commands()
  end
end

def _devices_registered()
  tasmota.remove_timer(_otaconf_timer_id)

  _initiate_file_transfer()
end

def _check_devices_registered()
  var devices = mesh.get_peers()
  var unregistered_devices = otaconf.get_unregistered_devices(devices)

  if unregistered_devices.size() > 0
    otaconf.register_devices(unregistered_devices, _devices_registered)
    tasmota.set_timer(3000, _otaconf_timeout, _otaconf_timer_id)
  else
    _initiate_file_transfer()
  end
end

def _scan_mesh()
  if _mesh_scan_retries > 0
    otaconf.send_mesh_scan_request()
    _mesh_scan_retries -= 1
    tasmota.set_timer(1000, _scan_mesh)
  else
    _check_devices_registered()
  end
end

def _process_config(response)
  tasmota.remove_timer(_otaconf_timer_id)

  if response != nil && response.contains('configuration')
    _configuration_uuid = response['uuid']
    _configuration_files = response['configuration']['files']
    _configuration_commands = response['configuration']['commands']

    if _configuration_files.size() > 0
      tasmota.log('OTA: attempting to fetch files', 2)

      for file : _configuration_files
        tasmota.log('OTA: fetching file: ' + file['uuid'], 2)

        if file.contains('url')
          tasmota.urlfetch(file['url'], file['uuid'])
        else
          tasmota.urlfetch(otaconf.get_server_url() + '/clusters/' + otaconf.get_cluster_uuid() + '/files/' + file['uuid'])
        end
      end
    end

    if response['configuration'].contains('mesh_scan') && response['configuration']['mesh_scan']
      _scan_mesh()
    else
      _initiate_file_transfer()
    end
  else
    _finish_otaconf(false)
  end
end

def _check_config_pending()
  otaconf.request_pending_config(_process_config)
  tasmota.set_timer(3000, _otaconf_timeout, _otaconf_timer_id)
end

def _cluster_registered(response)
  tasmota.remove_timer(_otaconf_timer_id)

  if response == nil
    _otaconf_failed()
  end

  _check_config_pending()
end

def _check_cluster_registered()
  if !otaconf.is_cluster_registered()
    otaconf.register_custer(_cluster_registered)
    tasmota.set_timer(3000, _otaconf_timeout, _otaconf_timer_id)
  else
    _check_config_pending()
  end
end

def _check_otaconf()
  var next_otaconf_time = int(persist.find('otaconf_last_poll')) + otaconf.get_poll_interval()
  var current_time = tasmota.rtc()['local']

  if !persist.has('otaconf_last_poll') || (current_time > next_otaconf_time) 
    _check_cluster_registered()
  else
    var remaining_hours = str((next_otaconf_time - current_time) / 3600)
    var remaining_min = str(((next_otaconf_time - current_time) % 3600) / 60)
    tasmota.log('OTA: next configuration poll in ~ ' + remaining_hours + ' hours ' + remaining_min + ' minutes', 2)
    _finish_otaconf(true)
  end
end

def _init_nbiot()
  print('init nb-iot') # todo implement
end

def _callback(sender, payload)
  var recieved_data = json.load(payload)
  print('receive payload: ' + payload) # todo remove
  print('receive size: ' + str(size(payload)))

  if recieved_data != nil && recieved_data.contains('action')
    var action = recieved_data['action']

    if action == 'TRANSFER_FILE'
      if recieved_data.contains('uuid')
        var uuid = recieved_data['uuid']

        if _devices_to_receive_file.contains(uuid)
          _devices_to_receive_file[uuid].push(sender)
        else
          _devices_to_receive_file[uuid] = [sender]
        end

        print(_devices_to_receive_file)
      end
    end
  elif payload != nil && size(payload) == _CHUNK_HEADER_SIZE
    var uuid = payload[0..11]
    var chunk_id = payload[12..14]

    print('chunks')
    print(chunk_id)
    print(_current_chunk_id)
    print('chunks')

    if re.match('^[0-9]*$', chunk_id) != nil
      print('yey')
      if int(chunk_id) == _current_chunk_id
        print('update')
        _current_chunk_id += 1
      end
      
      _send_file_chunk()
    end
  end
end

def _mesh_init()
  mesh.start()
  mesh.register_callback(_callback)
end

def _mesh_pre_init()
  mesh.init_soft_ap_mode(1)   
  
  tasmota.set_timer(50, _mesh_init)
end

def _wakeup_task()

  if !_wifi_used
    _init_nbiot()
    _mesh_pre_init()
  else 
    _mesh_init()
  end

  tasmota.set_timer(1000, _check_otaconf)
end

if !_wifi_used
  tasmota.add_rule('system#init', _wakeup_task)
else
  tasmota.add_rule('system#boot', _wakeup_task)
end

tasmota.add_cmd('OtaConfInterval', _set_ota_conf_interval)
tasmota.add_cmd('OtaServerUrl', _set_ota_server_url)
