import persist
import string
import mqtt
import mesh
import re

class ResponseHandler
  def handle_cluster_registered(response)
    if !response.contains('content') || !response['content'].contains('uuid')
      return false
    end

    persist.cluster_uuid = response['content']['uuid']
    persist.save()
    return true
  end

  def handle_devices_registered(response)
    if !response.contains('content') || response['content'].size() < 1 
      return false
    end

    if persist.has('registered_devices')
      var registered_devices = persist.find('registered_devices')
      
      for device : response['content']
        registered_devices.push(device)
      end

      persist.registered_devices = registered_devices
    else
      print(response['content'])
      persist.registered_devices = response['content']
    end

    persist.save()
    return true
  end

  def handle_response(payload)
    var response = json.load(payload)
  
    if response.contains('action')
      tasmota.log('OTA: receiving response with action: ' + response['action'], 2)
      var action = response['action']

      if action == 'REGISTER_CLUSTER'
        return self.handle_cluster_registered(response)
      elif action == 'REGISTER_DEVICES'
        return self.handle_devices_registered(response)
      elif action == 'POLL_CONFIGURATION'
        return true
      elif action == 'UPDATE_CONFIGURATION_STATUS'
        return true
      else
        tasmota.log('OTA: error handling response, unknown action', 1)
      end
    else
      tasmota.log('OTA: error handling response, invalid json', 1)
    end
  
    return false
  end
end

var otaconf = module('otaconf')

otaconf.init = def (m)
  class otaconf
    var mac
    var request_topic
    var response_topic
    var response_closure
    var subscribed
    var poll_interval
    var server_url
    var response_handler

    def _recv_response(cmd, idx, payload)
      if !self.response_handler.handle_response(payload)
        tasmota.log('OTA: error handling response, invalid content', 1)
        self.response_closure(nil)
      else
        self.response_closure(json.load(payload))
      end
    end

    def init()
      # self.mac = string.tolower(string.split(mesh.info()['MAC'], ':').concat())
      self.mac = nil
      self.request_topic = 'otaconf/%s/request'
      self.response_topic = 'otaconf/%s/response'
      self.response_closure = nil
      self.subscribed = false
      self.response_handler = ResponseHandler()

      if persist.has('otaconf_poll_interval')
        self.poll_interval = int(persist.find('otaconf_poll_interval'))
      else
        self.poll_interval = 86400 # 24 hours
      end

      if persist.has('otaconf_server_url')
        self.server_url = persist.find('otaconf_server_url')
      else
        self.server_url = ''
      end
    end

    def _send_request(request, closure)
      if !request.contains('action')
        tasmota.log('OTA: invalid request, abort sending', 1)
        return
      end

      tasmota.log('OTA: sending request with action: ' + request['action'], 2)

      self.response_closure = closure

      if !self.subscribed
        mqtt.subscribe(string.format(self.response_topic, self.get_mac()), def (cmd, idx, payload) self._recv_response(cmd, idx, payload) end)
        self.subscribed = true
      end

      mqtt.publish(string.format(self.request_topic, self.get_mac()), json.dump(request))
    end

    def get_poll_interval()
      return self.poll_interval
    end

    def set_poll_interval(interval)
      if re.match('^[0-9]*$', interval) != nil
        self.poll_interval = interval
        persist.otaconf_poll_interval = interval
        persist.save()
        
        return true
      else
        return false
      end
    end

    def get_server_url()
      return self.server_url
    end

    def set_server_url(url)
      self.server_url = url
      persist.otaconf_server_url = url
      persist.save()
    end

    def update_config_status(uuid, closure)
      tasmota.log('OTA: updating configuration status', 2)

      var cluster_uuid = persist.find('cluster_uuid')
      var content = {'status': 'finished'}
      var request = {'action': 'UPDATE_CONFIGURATION_STATUS', 'cluster_uuid': cluster_uuid, 'configuration_uuid': uuid, 'content': content}
      self._send_request(request, closure)
    end

    def request_pending_config(closure)
      tasmota.log('OTA: requesting pending configuration', 2)

      var cluster_uuid = persist.find('cluster_uuid')
      var request = {'action': 'POLL_CONFIGURATION', 'cluster_uuid': cluster_uuid, 'mac': self.get_mac()}
      self._send_request(request, closure)
    end

    def send_mesh_scan_request()
      tasmota.log('OTA: sending mesh scan request', 2)

      var request = {'action': 'MESH_SCAN'}
      mesh.broadcast(json.dump(request))
    end

    def get_unregistered_devices(devices)
      if devices == nil 
        return []
      end

      var registered_devices = persist.find('registered_devices')
      var unregistered_devices = []
      
      devices.push(self.get_mac())

      if registered_devices == nil
        return devices
      end

      for device : devices
        if registered_devices.find(device) == nil
          unregistered_devices.push(device)
        end
      end

      return unregistered_devices
    end

    def register_devices(devices, closure)
      tasmota.log('OTA: registering new devices', 2)

      var content = []

      for device : devices
        if device == self.get_mac()
          content.push({'mac': device, 'type': 'ESP32', 'gateway': true})
        else
          content.push({'mac': device, 'type': 'ESP32', 'gateway': false, 'peers': [self.get_mac(), device]})
        end
      end

      var cluster_uuid = persist.find('cluster_uuid')
      var request = {'action': 'REGISTER_DEVICES', 'cluster_uuid': cluster_uuid, 'content': content}
      self._send_request(request, closure)
    end

    def is_cluster_registered()
      return persist.has('cluster_uuid')
    end

    def get_cluster_uuid()
      return persist.find('cluster_uuid')
    end

    def register_custer(closure)
      tasmota.log('OTA: registering new cluster', 2)
      
      var request = {'action': 'REGISTER_CLUSTER'}
      self._send_request(request, closure)
    end

    def get_mac()
      if self.mac == nil
        self.mac = string.tolower(string.split(mesh.info()['MAC'], ':').concat())
        return self.mac
      else
        return self.mac
      end
    end

    def format_mac(mac)
      if size(mac) != 12
        return ''
      else
        return mac[0..1] + ':' + mac[2..3] + ':' + mac[4..5] + ':' + mac[6..7] + ':' + mac[8..9] + ':' + mac[10..11]
      end
    end
  end

  return otaconf()
end

return otaconf
